==> retry_toolkit-0.6.0-py3-none-any.whl: retry_toolkit/core.py <==
#┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅#
# SPDX-FileCopyrightText: © 2024 David E. James
# SPDX-License-Identifier: MIT
# SPDX-FileType: SOURCE
#┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
'''A class-based retry implementation.

Takes the same arguments as the "simple" version, but implements retry
as a class which is more easily extensible/modifiable.

Retry has been done and redone many times. Here is a version that takes only
a few arguments that are all optional but provides most of the flexibility
seen in many implementations.

It does not try to define all the different variables that may be used to
compute backoff values, instead preferring to allow a callable that could use
any algorithm desired to compute a backoff of which 3 very simple
implementations are provided. Users can use these as an example to setup their
own perfect backoff implementation (hopefully using jitter as well).

Or perhaps you should not use this module as a dependency, but instead copy
the strategy below, include it in your own codebase, and alter it to make it
your own. MIT is a permissive license.
'''
#┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅#
import functools
import time
from enum import Enum

from collections.abc import Callable

#┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
from .exceptions import (
    ExceptionTuple,
    ExceptionFunc,
    GiveUp,
)

from .defaults import (
    Defaults,
    LoggingDefaults,
)
from ._utils import (
    _ensure_callable,
    _process_subscriptions,
    _publish,
)
from .constants import (
    Events,
)


#──────────────────────────────────────────────────────────────────────────────#
# Decorator Factory
#──────────────────────────────────────────────────────────────────────────────#
def retry(
    tries      : int | Callable[[],int] = None,
    backoff    : int | Callable[[int],int] = None,
    exceptions : type(Exception) | ExceptionTuple | ExceptionFunc = None,
    class_f    = None,
    logger     = None,
    subscriptions = None,
    *args,
    **kwargs,
):
    _class_f = class_f or Defaults.RETRY_CLASS
    _class   = Retry if _class_f is None else _class_f()

    return _class(tries, backoff, exceptions, logger, subscriptions,
                    *args, **kwargs)


#──────────────────────────────────────────────────────────────────────────────#
# Retry Class
#──────────────────────────────────────────────────────────────────────────────#
class Retry:
    def __init__(self, tries, backoff, exceptions, logger, subscriptions,
                *args, **kwargs):
        self._tries      = tries
        self._backoff    = backoff
        self._exceptions = exceptions

        self._event_subs = _process_subscriptions(Events, subscriptions)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _result_is_success(self, result):
        return True

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _should_retry_exception(self, exc_type, exc_val):
        return True

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __iter__(self):
        '''starts a retry set'''
        self.iter_start = True

        self.target_func     = False
        self.func_successful = False

        self.done    = False
        self.try_num = 0
        self.n_tries = None
        self.total_sleep = 0
        self.exception_list = []

        return self

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __next__(self):
        '''starts a retry attempt'''
        if self.iter_start:
            self._event(Events.SETUP)
            self._setup()
            self._event(Events.START)
            self.iter_start = False

        if self.done:
            raise StopIteration()

        if self.n_tries == 0:
            self._event(Events.SKIP)
            raise StopIteration()

        if self.try_num < self.n_tries:
            return self

        self._giveup()

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __enter__(self):
        '''also called when starting a retry attempt'''
        if self.try_num > 0:
            self._sleep()
        self._event(Events.TRY)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __exit__(self, exc_type, exc_val, exc_tb):
        '''called at the end of a retry attempt with the exception if any'''
        func_success     = self.target_func is None and self.func_successful
        non_func_success = not self.target_func is None

        if exc_type is None:
            if func_success or non_func_success:
                self._event(Events.SUCCESS)
                self.done = True
            return True

        self._save_exception(exc_val)

        if not self._should_retry_exception(exc_type, exc_val):
            self._event(Events.ABORT)
            return False

        self._event(Events.FAIL)
        self._event(Events.FAIL_ON_EXCEPTION)

        self.try_num += 1
        return True

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = None
            for _try in self:
                with _try:
                    self.target_func     = func
                    self.func_successful = False

                    result = func(*args, **kwargs)

                    if self._result_is_success(result):
                        self.func_successful = True
                        break

                    self._event(Events.FAIL)
                    self._event(Events.FAIL_ON_RESULT)

            return result
        return wrapper

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _setup(self):
        # configure at call-time to allow any changes to defaults
        # to properly take effect each time func is used
        self._n_tries_f = _ensure_callable(self._tries      , Defaults.TRIES  )
        self._backoff_f = _ensure_callable(self._backoff    , Defaults.BACKOFF)
        self._exc_f     = _ensure_callable(self._exceptions , Defaults.EXC    )
        self._sleep_f   = Defaults.SLEEP_FUNC

        n_tries = self._n_tries_f()
        if n_tries < 0:
            self._warn(Warnings.NEGATIVE_TRIES)
            n_tries = 0

        self.n_tries = n_tries
        self.exc     = self._exc_f()

        # context/state
        self.total_sleep    = 0.0
        self.exception_list = []

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _sleep(self):
        sleep_time = self._backoff_f(self.try_num-1)

        if sleep_time < 0.0:
            self._warn(Warnings.NEGATIVE_SLEEP)
            sleep_time = 0.0

        self.total_sleep += sleep_time
        self._sleep_f(sleep_time)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _save_exception(self, e):
        self.exception_list.append(e)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _giveup(self):
        raise GiveUp(
            self.try_num+1,       # total tries
            self.total_sleep,     # total time sleeping (not total elapsed)
            self.target_func,     # function reference
            self.exception_list,  # all exceptions that happened
        )

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _event(self, event_id):
        context = dict(
            n_tries      = self.n_tries,
            try_num      = self.try_num,
            total_sleep  = self.total_sleep,
            target_func  = self.target_func,
            exceptions   = self.exception_list,
            #start_time   = self.start_time,
            #current_time = self.current_time,
        )
        _publish(event_id, self._event_subs, context)

==> retry_toolkit-0.6.1-py3-none-any.whl: retry_toolkit/core.py <==
#┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅#
# SPDX-FileCopyrightText: © 2024 David E. James
# SPDX-License-Identifier: MIT
# SPDX-FileType: SOURCE
#┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
'''A class-based retry implementation.

Takes the same arguments as the "simple" version, but implements retry
as a class which is more easily extensible/modifiable.

Retry has been done and redone many times. Here is a version that takes only
a few arguments that are all optional but provides most of the flexibility
seen in many implementations.

It does not try to define all the different variables that may be used to
compute backoff values, instead preferring to allow a callable that could use
any algorithm desired to compute a backoff of which 3 very simple
implementations are provided. Users can use these as an example to setup their
own perfect backoff implementation (hopefully using jitter as well).

Or perhaps you should not use this module as a dependency, but instead copy
the strategy below, include it in your own codebase, and alter it to make it
your own. MIT is a permissive license.
'''
#┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅┅#
import functools
import time
from enum import Enum

from collections.abc import Callable

#┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
from .exceptions import (
    ExceptionTuple,
    ExceptionFunc,
    GiveUp,
)

from .defaults import (
    Defaults,
    LoggingDefaults,
)
from ._utils import (
    _ensure_callable,
    PubSubManager,
    _get_logging_subscribers,
)
from .constants import (
    Events,
)


#──────────────────────────────────────────────────────────────────────────────#
# Decorator Factory
#──────────────────────────────────────────────────────────────────────────────#
def retry(
    tries      : int | Callable[[],int] = None,
    backoff    : int | Callable[[int],int] = None,
    exceptions : type(Exception) | ExceptionTuple | ExceptionFunc = None,
    class_f    = None,
    logger     = None,
    logging_config = None,
    subscriptions = None,
    *args,
    **kwargs,
):
    _class_f = class_f or Defaults.RETRY_CLASS
    _class   = Retry if _class_f is None else _class_f()

    return _class(tries, backoff, exceptions, logger, logging_config,
                  subscriptions,
                    *args, **kwargs)


#──────────────────────────────────────────────────────────────────────────────#
# Retry Class
#──────────────────────────────────────────────────────────────────────────────#
class Retry:
    def __init__(self, tries, backoff, exceptions, logger, logging_config,
                 subscriptions,
                *args, **kwargs):
        self._tries      = tries
        self._backoff    = backoff
        self._exceptions = exceptions

        self._pubsub = PubSubManager(Events, subscriptions)

        self._log_subs = _get_logging_subscribers(
                                    logging_config = logging_config,
                                    logger         = logger,
                                )

    def subscribe(self, event, func):
        return self._pubsub.subscribe(event, func)


    def unsubscribe(self, event, func):
        return self._pubsub.unsubscribe(event, func)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _result_is_success(self, result):
        return True

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _should_retry_exception(self, exc_type, exc_val):
        return True

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __iter__(self):
        '''starts a retry set'''
        self.iter_start      = True
        self.target_func     = False
        self.func_successful = False
        self.done            = False
        self.try_num         = 0
        self.n_tries         = None
        self.total_sleep     = 0
        self.exception_list  = []

        return self

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __next__(self):
        '''starts a retry attempt'''
        if self.iter_start:
            self._event(Events.SETUP)
            self._setup()
            self._event(Events.START)
            self.iter_start = False

        if self.done:
            raise StopIteration()

        if self.n_tries == 0:
            self._event(Events.SKIP)
            raise StopIteration()

        if self.try_num < self.n_tries:
            return self

        self._event(Events.GIVEUP)
        self._giveup()

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __enter__(self):
        '''also called when starting a retry attempt'''
        if self.try_num > 0:
            self._sleep()
        self._event(Events.TRY)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __exit__(self, exc_type, exc_val, exc_tb):
        '''called at the end of a retry attempt with the exception if any'''
        func_success     = self.target_func is None and self.func_successful
        non_func_success = not self.target_func is None

        if exc_type is None:
            if func_success or non_func_success:
                self._event(Events.SUCCESS)
                self.done = True
            return True

        self._save_exception(exc_val)

        if not self._should_retry_exception(exc_type, exc_val):
            self._event(Events.ABORT)
            return False

        self._event(Events.FAIL)
        self._event(Events.FAIL_ON_EXCEPTION)

        self.try_num += 1
        return True

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = None
            for _try in self:
                with _try:
                    self.target_func     = func
                    self.func_successful = False

                    result = func(*args, **kwargs)

                    if self._result_is_success(result):
                        self.func_successful = True
                        break

                    self._event(Events.FAIL)
                    self._event(Events.FAIL_ON_RESULT)

            return result
        return wrapper

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _setup(self):
        # configure at call-time to allow any changes to defaults
        # to properly take effect each time func is used
        self._n_tries_f = _ensure_callable(self._tries      , Defaults.TRIES  )
        self._backoff_f = _ensure_callable(self._backoff    , Defaults.BACKOFF)
        self._exc_f     = _ensure_callable(self._exceptions , Defaults.EXC    )
        self._sleep_f   = Defaults.SLEEP_FUNC

        n_tries = self._n_tries_f()
        if n_tries < 0:
            self._warn(Warnings.NEGATIVE_TRIES)
            n_tries = 0

        self.n_tries = n_tries
        self.exc     = self._exc_f()

        # context/state
        self.total_sleep    = 0.0
        self.exception_list = []

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _sleep(self):
        sleep_time = self._backoff_f(self.try_num-1)

        if sleep_time < 0.0:
            self._warn(Warnings.NEGATIVE_SLEEP)
            sleep_time = 0.0

        self.total_sleep += sleep_time
        self._sleep_f(sleep_time)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _save_exception(self, e):
        self.exception_list.append(e)

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _giveup(self):
        raise GiveUp(
            self.try_num+1,       # total tries
            self.total_sleep,     # total time sleeping (not total elapsed)
            self.target_func,     # function reference
            self.exception_list,  # all exceptions that happened
        )

    #┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈┈#
    def _event(self, event_id):
        context = dict(
            n_tries      = self.n_tries,
            try_num      = self.try_num,
            total_sleep  = self.total_sleep,
            target_func  = self.target_func,
            exceptions   = self.exception_list,
            #start_time   = self.start_time,
            #current_time = self.current_time,
        )
        self._pubsub.publish(event_id, context)
